import logging
import openai
from tenacity import retry, retry_if_exception_type, stop_after_attempt
from tenacity import wait_random_exponential, before_sleep_log
import traceback

token_error_message = 'tokens for the input and instruction but the maximum allowed is 3000. ' \
                      'Please reduce the input or instruction length.'


@retry(retry=retry_if_exception_type(openai.error.APIError) |
             retry_if_exception_type(openai.error.APIConnectionError) |
             retry_if_exception_type(openai.error.ServiceUnavailableError),
       wait=wait_random_exponential(max=300),
       stop=stop_after_attempt(5),
       before_sleep=before_sleep_log(logging.getLogger(), logging.ERROR))
@retry(retry=retry_if_exception_type(openai.error.RateLimitError),
       wait=wait_random_exponential(max=600),
       # We want the experiment to crash in this situation
       # We've launched too many
       stop=stop_after_attempt(5),
       before_sleep=before_sleep_log(logging.getLogger(), logging.INFO))
def query_gpt(source=None, instruction=None, modality='code', n=1, t=1.0):
    """
    Get code snippets from GPT-3.

    If instruction is not specified, the code is extended (autocompleted),
    otherwise it's edited according to the instruction.
    """
    logging.info(f"Querying GPT with temperature {t} and {n} snippets.")

    result = []
    try:
        if instruction:
            if modality == 'text':
                engine = "text-davinci-edit-001"
                logging.info("Calling GPT for text editing")

            elif modality == 'code':
                engine = "code-davinci-edit-001"
                logging.info("Calling GPT for code editing")
            else:
                raise ValueError(f'Unknown modality: {modality}')

            response = openai.Edit.create(
                engine=engine,
                input=source,
                n=n,
                instruction=instruction,
                temperature=t
            )
            result = [choice['text'] for choice in response["choices"]
                      if "text" in choice.keys()]
        else:
            if modality == 'text':
                engine = "text-davinci-003"
                logging.info("Calling GPT for text completion")
            elif modality == 'code':
                engine = "code-davinci-002"
                logging.info("Calling GPT for code completion")
            else:
                raise ValueError(f'Unknown modality: {modality}')

            response = openai.Completion.create(
                engine=engine,
                prompt=source,
                n=n,
                temperature=t
            )

            if modality == 'text':
                result = [choice['text'] for choice in response["choices"]
                          if "text" in choice.keys()]
                logging.info(f"\nBug summary by GPT:\n{result[0]}")
            else:
                result = [source + '\n' + choice['text'] for choice in response["choices"]
                          if "text" in choice.keys()]

    except openai.error.InvalidRequestError as e:
        result = []
        if token_error_message in e.error.message:
            raise e

    return result

def explore_gpt(source='', instruction=None, modality='code', batch_size=1, 
                t=0.0, delta_t=0.2, log_gpt_call=lambda **kwargs: None):
    """Get many code snippets from GPT-3 ordered from most to least likely"""

    # Beam search would be preferable, but it's computationally costly
    # (for OpenAI, which is why they don't offer it)

    while t <= 1:
        try:
            log_gpt_call(source=source, instruction=instruction, modality=modality,
                         n=batch_size, t=t)
            yield from query_gpt(source=source, instruction=instruction, modality=modality,
                                 n=batch_size, t=t)
        except openai.error.Timeout as e:
            logging.info(traceback.format_exc())
            pass
        except openai.error.InvalidRequestError as e:
            logging.error(traceback.format_exc())

            if token_error_message in e.error.message:
                logging.info('Stopping iterations due to token limit error')
                break

        t += delta_t

if __name__ == '__main__':
    import itertools
    logging.basicConfig(level=logging.INFO)

    for code in itertools.islice(explore_gpt(instruction=input()), 10):
        print(code)



import itertools
import logging

from seidr.gpt import explore_gpt
from seidr.prompt import initial_prompt, write_debug_prompt, start_coding


def rolling_best(objects, max_score=1, metric=lambda x: x):
    best_score = None

    for object in objects:
        score = metric(object)
        if best_score is None or score > best_score:
            best_score = score
            try:
                logging.info(f'\nThe program has improved. Code: \n\n{object.read()}\n\n')
            except:
                pass
            yield object

        if best_score >= max_score:
            break


def beam_search(beam, update, metric, beam_width=100):
    """Generic evolutionary algorithm for improving anything"""

    new_beam = []

    # yield beam_width draft (non-repaired) programs
    for code in beam:
        yield code
        new_beam.append(code)

    while True:
        beam = sorted(new_beam, key=metric, reverse=True)[:beam_width]
        if len(beam) == 0:
            break
        new_beam = []

        # yield beam_width * branching_factor children (repaired programs)
        for parent in beam:
            for child in update(parent):
                yield child
                new_beam.append(child)


def distribute_heat(heat, n, batch_size):
    if n == 1:
        t = 0
        delta_t = 0.2  
    else:
        # We intentionally avoid temperature=0
        # That would lead to a batch of identical code snippets
        # Update temperature but keep it 1 at max
        batch_count = n // batch_size + 1
        delta_t = heat / batch_count
        t = delta_t

    return t, delta_t


def draft(task_description, examples, language, batch_size=10, limit_n=None, 
          log_gpt_call=lambda **kwargs: print(kwargs)):
    t, delta_t = distribute_heat(1, limit_n, batch_size)
        
    prompt = initial_prompt(task_description, examples)
    start = start_coding(prompt, language=language)
    codes = explore_gpt(source=start, instruction=task_description, modality='code',
                        batch_size=batch_size, 
                        t=t, 
                        delta_t=delta_t,
                        log_gpt_call=log_gpt_call)

    if limit_n:
        codes = itertools.islice(codes, limit_n)

    return codes


def debug(code, debug_prompt_text, n, batch_size=10, log_gpt_call=print):
    """Generate n attempts to fix program so that it passes tests"""
    t, delta_t = distribute_heat(1, n, batch_size)

    codegen = explore_gpt(source=code,
                          instruction=debug_prompt_text,
                          modality='code',
                          batch_size=batch_size,
                          t=t, delta_t=delta_t,
                          log_gpt_call=log_gpt_call)
    return itertools.islice(codegen, n)

def print_code(code, **vars):
    print(vars)
    print(code)

def develop(task_description, 
            examples,
            critics,
            language='C++',
            beam_width=3,
            branching_factor=10,
            max_programs=None,
            log_metrics=print,
            log_attempt=print_code,
            log_solution=lambda *args, **kwargs: print('This program is the best!'),
            log_gpt_call=lambda *args, **kwargs: print(kwargs),
            batch_size=10):
    """
    Write a program in language that solves task and passes tests.
    Solve debug-rewrite trade-off with beam search of given beam size

    examples is a sequence of (inputs, outputs) pairs
    where inputs and outputs are sequences of strings (lines of code)
    likewise for tests

    examples are used in the prompt for the language model,
    while tests are used to select the best solution

    Returns a generator of programs where each program passes
    more tests than the previous one. The last program in the generator
    passes all tests.
    """

    def have_kids(candidate):
        logging.debug(f'Running debug_and_test')
        prompt, code, evals = candidate
        worst_eval = min(evals, key=lambda e: e.score())
        feedback = worst_eval.pen_report()
        
        for code in debug(code, feedback,
                          n=branching_factor, batch_size=batch_size,
                          log_gpt_call=log_gpt_call):
            yield feedback, code, [critic(code) for critic in critics]

    def metric(candidate):
        prompt, code, evals = candidate
        avg_score = sum(e.score() for e in evals) / len(evals)
        return avg_score
    
    beam = draft(task_description, examples, language, batch_size=batch_size, 
                 limit_n=beam_width, log_gpt_call=log_gpt_call)
    beam = [(task_description, code, [critic(code) for critic in critics])
            for code in beam]
    
    best_score = float('-inf')

    search = beam_search(beam, have_kids, metric, beam_width)
    for idx, candidate in enumerate(search):
        prompt, code, evals = candidate

        avg_score = sum(e.score() for e in evals) / len(evals)
        test_pass_rate = sum(e.check() for e in evals) / len(evals)

        logging.info(f'Current program:\n{code}')

        metrics = {
            'idx': idx,
            'avg_score': avg_score,
            'pass_rate': test_pass_rate
        }

        log_metrics(metrics)
        log_attempt(code, idx=idx, 
                    prompt=prompt, test_pass_rate=test_pass_rate)

        if avg_score > best_score:
            best_score = avg_score
            log_metrics({f'best_{metric}': val for metric, val in metrics.items()})
            log_solution(code, idx=idx, 
                         prompt=prompt, test_pass_rate=test_pass_rate)

            if test_pass_rate == 1:
                break

        if max_programs is not None and (idx == max_programs - 1):
            break

    return code


if __name__ == '__main__':
    from seidr.eval import IOMatch

    task = 'A program that outputs "Hello World"'

    # Use the same IO examples for prompt and tests
    critics = [
        lambda code: IOMatch(code, language='Python', 
                             input=[''], output=['Hello World'])
    ]
    develop(task, [[[''], ['Hello, World']]], critics, language='Python')



"""
Handy tools for recording our generated programs to a git repository.
May be nice to include this into programlib one day
"""

import git
import gitdb
from git import GitError, NoSuchPathError
from git import Repo

from tenacity import retry, retry_if_exception_type, stop_after_attempt
from tenacity import wait_random_exponential
from programlib import Program

from pathlib import Path
from uuid import uuid4
import os
import shutil
import logging

@retry(retry=retry_if_exception_type(git.exc.GitCommandError) |
             retry_if_exception_type(git.exc.BadName) |
             retry_if_exception_type(gitdb.exc.BadName),
       wait=wait_random_exponential(),
       stop=stop_after_attempt(50))
def pullpush(repo):
    repo.remotes.origin.pull()
    repo.remotes.origin.push()


def upload_file(repo, filename, message=None):
    if not message:
        message = f'added {filename}'

    pullpush(repo)
    repo.index.add(filename)
    repo.index.commit(message)
    pullpush(repo)
    commit_url = repo.remotes.origin.url + "/commit/" + repo.head.commit.hexsha
    commit_url = 'https://' + commit_url[commit_url.find('@github.com') + 1:]
    logging.info(f'Tried to push updates to git repo.\n'
                 f'See commit at {commit_url}. \n'
                 f'Commit message: {message}')


def ensure_repo(remote, path, branch=None):
    try:
        repo = Repo(path)

        if branch:
            repo.git.checkout(branch)
    except GitError as e:
        logging.info(f'Git error in ensure repo {e}')
        shutil.rmtree(path, ignore_errors=True)
        repo = Repo.clone_from(remote, path)

        if branch:

            repo.git.fetch('--all')
            branches = [ref.name for ref in repo.references]

            if f'{repo.remote().name}/{branch}' in branches:
                repo.git.checkout(branch)
            else:
                repo.git.checkout('-b' + branch)
                repo.git.add(repo.working_dir)
                repo.git.push('--set-upstream', repo.remote().name, branch)

    return repo


def config_repo(dir, branch):
    try:
        import os
        logging.info('config_repo')
        repo = ensure_repo(os.environ['GIT_REMOTE'], dir, branch=branch)
        repo.config_writer().set_value('user', 'name', os.environ['GIT_USER']).release()
        repo.config_writer().set_value('user', 'email', os.environ['GIT_EMAIL']).release()
        repo.config_writer().set_value('pull', 'rebase', False).release()
        return repo
    except (KeyError, GitError) as e:
        logging.info(f'config_repo exception {e}')
        return None

class FileLogger:
    """
    A careful archivist keeping track of your program's versions.
    If git environment variables are set, will synchronize with a remote repository.
    """

    def __init__(self, branch, filename, commit_msg_template='{message}'):
         os.makedirs('solutions', exist_ok=True)
         self.filename = filename
         self.commit_msg_template = commit_msg_template
         cache_dir = branch + '_' + str(uuid4())[:6]
         self.dir = Path('solutions') / cache_dir
         self.repo = config_repo(self.dir, branch=branch)
         os.makedirs(self.dir, exist_ok=True)

    # def current(self):
    #     return Program(workdir=self.dir,
    #                    name=self.filename,
    #                    language=self.language)

    def log(self, content, **vars):
        with open(self.dir / self.filename, 'w') as f:
            f.write(content)
        if self.repo:
            commit_msg = self.commit_msg_template.format(**vars)
            upload_file(self.repo, self.filename, commit_msg)

    def __call__(self, *args, **kwargs):
        return self.log(*args, **kwargs)


import logging
from pathlib import Path
from string import Template

from programlib import language_

from seidr.gpt import explore_gpt
from seidr import get_template

dont_change = 'Do not change anything'

def initial_prompt(task_description, examples):
    prompt = task_description
    prompt += '\nFor example,'
    for sample_inputs, sample_outputs in examples:
        prompt += ''
        prompt += '\ninput:\n'
        for sample_input in sample_inputs:
            prompt += sample_input + '\n'
        prompt += 'output:'
        for sample_output in sample_outputs:
            prompt += '\n' + sample_output
    return prompt

def gpt_assisted_prompt(debug_prompt_text, task_description, input, expected_output, actual_output):
    """
    Create description of a bug using GPT3 completion.
    """
    assert 'GPT ---' in debug_prompt_text and len(debug_prompt_text.split(' --- ')) >= 3, 'Invalid prompt'
    _, code_behavior, debug_prompt_text = debug_prompt_text.split(' --- ')
    # Form problem description using template
    code_behavior = code_behavior.format(
        t=task_description,
        i=input,
        o=expected_output,
        a=actual_output)

    # Get GPT summary of a bug
    logging.info(f'\nGPT summary of a bug:\n{code_behavior}')
    bug_description = next(explore_gpt(source=code_behavior, instruction=None,
                                       modality='text', batch_size=1, t=0.0, delta_t=0.2))
    # Form debug prompt using template
    debug_prompt_text = debug_prompt_text.format(
        s=bug_description,
        t=task_description,
        i=input,
        o=expected_output,
        a=actual_output)
    return debug_prompt_text


def write_debug_prompt(mistake, debug_prompt_text, task_description=None):
    logging.info('Updating debug prompt')
    output_lines = '\n'.join([s.decode("utf-8") if type(s) == bytes else s for s in mistake.output_lines])

    prompt = ""
    if mistake.correctness < 1:
        if mistake.exit_status:
            error_lines = output_lines
            prompt = f'Fix {error_lines}'

        else:
            i = '\\n'.join(mistake.input_lines)
            o = '\\n'.join(mistake.expected_output_lines)
            if 'GPT ---' in debug_prompt_text:
                
                prompt = gpt_assisted_prompt(
                    debug_prompt_text, task_description, mistake.input_lines,
                    mistake.expected_output_lines, output_lines)
            else:
                prompt = debug_prompt_text.format(i=i, o=o)
    else:
        logging.info('\n\nrun.correctness = 1 for all runs, mistake lines are empty\n\n')
        prompt = dont_change

    logging.info(f'The prompt is: \n{prompt}')
    return prompt


def start_coding(prompt, language='C++'):
    language = language_(language)
    template = language.source.format(name='template')
    template = get_template(template)
    template = Template(template)
    return template.substitute(prompt=prompt)



from abc import ABC
from programlib import Program
from seidr.prompt import write_debug_prompt, dont_change

class Evaluation(ABC):
    """
    A method for evaluating a system
    Produces a binary pass/fail result, a float score, and a text report
    """

    def __init__(self, SUT, passing_score=1):
        """
        SUT: System Under Test
        passing_score: float score required to pass the evaluation
        """

        self.SUT = SUT
        self.passing_score = passing_score

    def check(self):
        """Produce a binary pass/fail evaluation"""
        return self.score() >= self.passing_score

    def score(self):
        """Produce a float score"""
        pass

    def pen_report(self):
        """Produce a text report"""
        pass

class IOMatch(Evaluation):
    def __init__(self, code, language, input, output, 
                 debug_template='Make sure {i} -> {o}', 
                 task_description=None):
        program = Program(code, language=language)
        super().__init__(program)
        self.input = input
        self.output = output
        self.test_run = None
        self.debug_template = debug_template
        self.task_description = task_description

    def run_test(self, rerun=True):
        if rerun or not self.test_run:
            self.test_run = self.SUT.test([[self.input, self.output]])[0]

    def score(self):
        self.run_test(rerun=False)
        return self.SUT.avg_score

    def pen_report(self):
        self.run_test(rerun=False)
        return write_debug_prompt(self.test_run, 
                                  self.debug_template, 
                                  self.task_description)

class UnitTest(Evaluation):
    def __init__(self, code, language, test):
        program = Program(code + '\n' + test, language=language)
        super().__init__(program)
        self.output = None

    def run_test(self, rerun=True):
        if rerun or not self.output:
            self.output = self.SUT.run(force=True)

    def score(self):
        self.run_test(rerun=False)
        return not self.SUT.exit_status

    def pen_report(self):
        self.run_test(rerun=False)
        if self.score():
            return dont_change
        else:
            return self.output